from pydantic import BaseModel
from typing import Optional
from functools import partial
from .const import REG_TYPE

from box import Box

class Callbacks(BaseModel):
    holding_reg: partial
    input_reg: partial
    write_reg: partial

    class Config:
        arbitrary_types_allowed = True

class Register(BaseModel):
    addr: int
    value: Optional[float]=None
    reg_type: REG_TYPE

class InputRegister(Register):
    reg_type: REG_TYPE = REG_TYPE.INPUT


class HoldingRegister(Register):
    reg_type: REG_TYPE = REG_TYPE.HOLDING

## https://shop.systemair.com/upload/assets/SAVE_MODBUS_VARIABLE_LIST_20190116__REV__29_.PDF
## registry values - 1 for some reason







from box import Box

from .models import (
     InputRegister
    ,HoldingRegister
)

registers = Box({
    #Input registers
     "sa_target_temperature_"      : InputRegister(addr=2053) # REG_TC_SP_SATC: Temperature setpoint for the supply air temperature
    ,"remaining_filter_time_"      : InputRegister(addr=7004) # REG_FILTER_REMAINING_TIME_L: Remaining filter time in seconds, lower 16 bits
    ,"heat_exchanger_"              : InputRegister(addr=14102) # REG_OUTPUT_Y2_ANALOG: Heat Exchanger AO state.
    ,"heat_exchanger_active_"        : InputRegister(addr=14103) # REG_OUTPUT_Y2_DIGITAL: Heat Exchanger DO state.0: Output not active1: Output active
    # heater_a0                 : InputRegister(addr=14100)# REG_OUTPUT_Y1_ANALOG: Heater AO state.
    ,"heater_active_"                 : InputRegister(addr=14101) # REG_OUTPUT_Y1_DIGITAL: Heater DO state:0: Output not active1: Output active
    ,"filter_alarm_"               : InputRegister(addr=7006) # REG_FILTER_ALARM_WAS_DETECTED: Indicates if the filter warning alarm was generated.
    ,"usermode_"                   : InputRegister(addr=1160) # REG_USERMODE_MODE: Active User mode.0: Auto1: Manual2: Crowded3: Refresh4: Fireplace5: Away6: Holiday7: Cooker Hood8: Vacuum Cleaner9: CDI110: CDI211: CDI312: PressureGuard
    # speed_saf_desired_off     : InputRegister(addr=1351) # REG_SPEED_SAF_DESIRED_OFF: Indicates that the SAF shall be turned off once the electrical reheater is cooled down
    ,"saf_speed_"                  : InputRegister(addr=14370)  # REG_OUTPUT_FAN_SPEED1: Supply air fan control signal in %
    ,"eaf_speed_"                  : InputRegister(addr=14371)  # REG_OUTPUT_FAN_SPEED2: Extract air fan control signal in %
    # speed_saf_desired_off     : InputRegister(addr=1351) 
    # speed_saf_desired_off     : InputRegister(addr=1351) 
    # speed_saf_desired_off     : InputRegister(addr=1351) 
    
    #HoldingRegister
    ,"target_temperature"          : HoldingRegister(addr=2000)   # REG_TC_SP: Temperature setpoint for the supply air temperature
    ,"sa_temperature_sensor"       : HoldingRegister(addr=12102)   # REG_SENSOR_SAT: Supply Air Temperature sensor (standard)
    ,"oa_temperature_sensor"       : HoldingRegister(addr=12100)   # REG_SENSOR_OAT: Outdoor Air Temperature sensor (standard)
    ,"saf_usermode_fs"             : HoldingRegister(addr=1130)    # REG_USERMODE_MANUAL_AIRFLOW_LEVEL_SAF:  Fan speed level for mode Manual, supply fan.(1): value Off only allowed if contents of register 1352 is 1: Off 2: Low 3: Normal 4:High
    ,"eaf_usermode_fs"             : HoldingRegister(addr=1131)   # REG_USERMODE_MANUAL_AIRFLOW_LEVEL_EAF Fan speed level for mode Manual, extract fan. 2: Low 3: Normal 4: High
    ,"usermode"                    : HoldingRegister(addr=1161)   # REG_USERMODE_HMI_CHANGE_REQUEST: New desired user mode as requested bMI0: None1: AUTO2: Manual 3: Crowded4:Refresh5: Fireplace6: Away7: Holiday
    ,"pdm_humidity_sensor"         : HoldingRegister(addr=12135)   # REG_SENSOR_RHS_PDM: PDM RHS sensor value (standard)
    # REG_SENSOR_PDM_EAT_VALUE  : = HoldingRegister(addr=12543)  # REG_SENSOR_PDM_EAT_VALUE: PDM EAT sensor value (standard)
    # REG_TC_CASCADE_SP_MIN     : = HoldingRegister(addr=2020)  # REG_TC_CASCADE_SP_MIN: Minimum temperature set point for the SATC
    # REG_TC_CASCADE_SP_MAX     : = HoldingRegister(addr=2021)  # REG_TC_CASCADE_SP_MAX: Maximum temperature set point for the SATC
    ,"target_humidity"             : HoldingRegister(addr=2202) #REG_ROTOR_RH_TRANSFER_CTRL_SETPOINT: Set point setting for RH transfer contro
    ,"co2_sensor"                  : HoldingRegister(addr=12115) # REG_SENSOR_CO2S: CO2 value (accessory)
    ,"cooling_recovery_limit"      : HoldingRegister(addr=2314) # REG_COOLER_RECOVERY_LIMIT_T:Temperature at which cooling recovery is allowed
    ,"cooling_recovery_enabled"    : HoldingRegister(addr=2133)  #REG_HEAT_EXCHANGER_COOLING_RECOVERY_ON_OFF: Enabling of cooling recovery
    ,"free_cooling_enabled"        : HoldingRegister(addr=12135)  #REG_FREE_COOLING_ON_OFF: Indicates if free cooling is enabled
    ,"free_cooling_start_time_h"   : HoldingRegister(addr=4105) # REG_FREE_COOLING_START_TIME_H: Start time of free cooling night-period, hour.Valid range is from 0 to 8 and from 21 to 23.
    ,"free_cooling_start_time_m"   : HoldingRegister(addr=4106) # REG_FREE_COOLING_START_TIME_M: 0 59 Start time of free cooling night-period,Minute
    ,"free_cooling_end_time_h"     : HoldingRegister(addr=4107) # REG_FREE_COOLING_END_TIME_H: End time of free cooling night-period, hour.Valid range is from 0 to 8 and from 21 to 23.
    ,"free_cooling_end_time_m"     : HoldingRegister(addr=4108) # REG_FREE_COOLING_END_TIME_M: 0 59 End time of free cooling night-period, Minute
    ,"free_cooling_active"         : HoldingRegister(addr=4110) # REG_FREE_COOLING_ACTIVE: 0 1 Indicates if free cooling is being performed
    # humidity                    = HoldingRegister(addr=12135) 
    # humidity                    = HoldingRegister(addr=12135) 
    # humidity                    = HoldingRegister(addr=12135) 
    # humidity                    = HoldingRegister(addr=12135) 
    # humidity                    = HoldingRegister(addr=12135)         
    # humidity                    = HoldingRegister(addr=12135) 

})


from typing import Awaitable, Callable
from functools import partial
import logging
_LOGGER = logging.getLogger(__name__)

from .const import *
from .models import Callbacks, Register
from .registers import registers



class PySystemAir():

    _async_callbacks: Callbacks

    def __init__(self, async_callbacks:Callbacks=None):
        _LOGGER.info(f"New SYSTEMAIR MODULE")
        if async_callbacks is not None:
            self._async_callbacks=async_callbacks
        else:
            from pymodbus.client.asynchronous.serial import AsyncModbusSerialClient as Client
            event_loop, client = Client(port='/dev/ttyUSB0',
                      stopbits=1,
                      bytesize=8,
                      parity='E',
                      baudrate=56000,
                      timeout=2)
            client.connect()
            # client.read_holding_registers()
            # holding_reg_callback=partial(client.read_holding_registers, 
            self._async_callbacks = Callbacks(holding_reg=client.read_holding_registers, input_reg=client.read_input_registers, write_reg=client.write_register)
            # callbacks_accepted=True
            # if iscoroutinefunction(async_callbacks['holding_reg']):
            #     self._async_callbacks.holding_reg=async_callbacks['holding_reg']
            # else
            # self._async_callbacks.input_reg=async_callback_input_reg
            # self._async_callbacks.write_reg=async_callback_write_reg

        # registers = RegMap


    async def async_update_all(self):
        """
        Updates all of the input and holding regs dict values.
        """
       
        for key, register in registers.items():
            # _LOGGER.warning(f"Updating register for {key} {register}")
            # _LOGGER.warning(f"register.addr == {register.addr}")
            # _LOGGER.warning(f"register.reg_type == {register.reg_type}")
            try:
                await self.async_update_from_register(key, register)
                # result=None
                # if register.reg_type == REG_TYPE.INPUT:
                #     result = await self._async_callbacks.input_reg(address=register.addr)
                # elif register.reg_type == REG_TYPE.HOLDING:
                #     result = await self._async_callbacks.holding_reg(address=register.addr)
                # else:
                #     _LOGGER.warning(f"register.reg_type not matched")
                # # _LOGGER.warning(f"result= {result}")
                # if result is None:
                #     _LOGGER.warning(f"Error reading {variable} value from SystemAir modbus adapter")
                # else:
                #     register.value = result.registers[0]
            except AttributeError:
                 _LOGGER.warning(f"Modbus read failed for {key}")

    async def async_update_from_register(self, key, register:Register):
        """
        Updates all of the input and holding regs dict values.
        """
        try:
            result=None
            _LOGGER.warning(f"register.addr == {register.addr}")
            _LOGGER.warning(f"register.reg_type == {register.reg_type}")
            if register.reg_type == REG_TYPE.INPUT:
                _LOGGER.warning(f"calling {self._async_callbacks.input_reg}")
                result = await self._async_callbacks.input_reg(address=register.addr)
            elif register.reg_type == REG_TYPE.HOLDING:
                _LOGGER.warning(f"calling {self._async_callbacks.holding_reg}")
                result = await self._async_callbacks.holding_reg(address=register.addr)

            if result is None:
                 _LOGGER.warning(f"Error reading {key} value from SystemAir modbus adapter")
            else:
                _LOGGER.warning(f"{key} value is {result.registers[0]}")
                register.value = result.registers[0]
        except AttributeError as e:
            raise e


    async def async_write_to_register(self, register:Register, value):
        """
        Updates all of the input and holding regs dict values.
        """
        try:
            if await self._async_callbacks.write_reg(address=register.addr, value=value):
                    register.value = value
            else:
                _LOGGER.error(f"Unable to write {value} to register {register.addr}") 
        except AttributeError as e:
            raise e

    # @property
    # def user_modes(self):
    #     """Return the fan setting."""
    #     return list( USER_MODES.values())




    @property
    def fan_mode(self):
        """Return the fan setting."""
        if registers.saf_usermode_fs.value is None:
            return 3
        return registers.saf_usermode_fs.value



    @property
    def filter_hours(self):
        if registers.remaining_filter_time_.value is None:
            return 0
        return registers.remaining_filter_time_.value 

    @property
    def filter_alarm(self):
        if registers.filter_alarm_.value is None:
            return 0
        return registers.filter_alarm_.value

    @property
    def heat_exchanger_active(self):
        if registers.heat_exchanger_active_.value is None:
            return 0
        return registers.heat_exchanger_active_.value
    
    @property
    def heat_exchanger(self):
        if registers.heat_exchanger_.value is None:
            return 0
        return registers.heat_exchanger_.value

 
    # @property
    # def heating(self):
    #     if registers.oa_temperature_sensor.value is None:
    #         return 0
    #     return registers.oa_temperature_sensor.value / 10.0


    @property
    def heater_enabled(self):
        if registers.heater_active_.value is None:
            return 0
        return registers.heater_active_.value

    # @property
    # def cooling_enabled(self):  
    #     if registers.oa_temperature_sensor.value is None:
    #         return 0
    #     return registers.oa_temperature_sensor.value / 10.0


    @property
    def free_cooling_enabled(self): 
        if registers.free_cooling_enabled.value is None:
            return 0
        return registers.free_cooling_enabled.value


    @property
    def free_cooling_active(self):
        if registers.free_cooling_active.value is None:
            return 0
        return registers.free_cooling_active.value


    @property
    def free_cooling_start_time(self):
        if registers.free_cooling_start_time_h.value is None or registers.free_cooling_start_time_m.value is None:
            return 0
        return f"{registers.free_cooling_start_time_h.value}:{registers.free_cooling_start_time_m.value}"

    @property
    def free_cooling_end_time(self): 
        if registers.free_cooling_end_time_h.value is None or registers.free_cooling_end_time_m.value is None:
            return 0
        return f"{registers.free_cooling_end_time_h.value}:{registers.free_cooling_end_time_m.value}"

    @property
    def outdoor_air_temp(self):
        if registers.oa_temperature_sensor.value is None:
            return 0
        return registers.oa_temperature_sensor.value / 10.0


    @property
    def target_temperature(self):
        """Return the temperature we try to reach."""
        if registers.target_temperature.value is None:
            return 0
        return registers.target_temperature.value / 10.0

    @property
    def current_temperature(self):
        """Return the current temperature."""
        if registers.sa_temperature_sensor.value is None:
            return 0
        return registers.sa_temperature_sensor.value / 10

    @property
    def current_humidity(self):
        """Return the temperature we try to reach."""
        if registers.pdm_humidity_sensor.value is None:
            return 0
        return registers.pdm_humidity_sensor.value

    @property
    def target_humidity(self):
        """Return the temperature we try to reach."""
        if registers.target_humidity.value is None:
            return 0
        return registers.target_humidity.value

    @property
    def current_co2_ppm(self):
        """Return the current_co2_ppm meas."""
        if registers.co2_sensor.value is None:
            return 0
        return registers.co2_sensor.value


    async def async_set_temperature(self, value):
        """Set new target temperature."""
        await self.async_write_to_register(registers.target_temperature, value)


    async def async_set_fan_mode(self, value):
        """Set new fan mode."""
        value = FAN_MODES(value)
        await self.async_write_to_register(registers.saf_usermode_fs, value)
        await self.async_write_to_register(registers.eaf_usermode_fs, value)


    async def async_set_humidity(self, value):
        """Set new target temperature."""
        value=int(round(value))
        await self.async_write_to_register(registers.target_humidity, value)
        

    async def async_cotwo_meas(self, value):
        """Set new target temperature."""
        await self.async_write_to_register(registers.co2_sensor, value)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysystemair']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0', 'pymodbus>=2.5.3,<3.0.0', 'python-box>=6.0.2,<7.0.0']

setup_kwargs = {
    'name': 'pysystemair',
    'version': '0.3.6',
    'description': 'Python package to interface a SystamAir SAVE VTR ventilation through modbus. Heavily insipired by pyflexit (https://github.com/Sabesto/pyflexit). Made with the purpose of working with a home assistant component',
    'long_description': None,
    'author': 'mvheimburg',
    'author_email': 'heimburg@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Constants for the systemair integration."""

from typing import Final
from enum import IntEnum, Enum


# DOMAIN = "systemair"
DEVICE_DEFAULT_NAME = "SystemAir"
SAVE_VTR = "save_vtr"

UPDATE_ON_READ = True
DEFAULT_TEMPERATURE = 20

class REG_TYPE(IntEnum):
    INPUT=1
    HOLDING=2

class FAN_MODES(IntEnum):
    OFF=1
    LOW=2   
    NORMAL=3
    HIGH=4


SA_OPERATION_MODE_AUTO = "auto"
SA_OPERATION_MODE_MANUAL = "manual"
SA_OPERATION_MODE_CROWDED = "crowded"
SA_OPERATION_MODE_REFRESH = "refresh"
SA_OPERATION_MODE_FIREPLACE = "fireplace"
SA_OPERATION_MODE_HOLIDAY = "holiday"
SA_OPERATION_MODE_IDLE = "idle"
SA_OPERATION_MODE_OFF = "off"

USER_MODES = {
    0: SA_OPERATION_MODE_AUTO,
    1: SA_OPERATION_MODE_MANUAL,
    2: SA_OPERATION_MODE_CROWDED,
    3: SA_OPERATION_MODE_REFRESH,
    4: SA_OPERATION_MODE_FIREPLACE,
    5: SA_OPERATION_MODE_IDLE,
    6: SA_OPERATION_MODE_HOLIDAY,
}

ATTR_HUMIDITY: Final = "humidity"